package com.ijchart.xychart.render.category ;

import java.awt.Color ;
import java.awt.Paint ;
import java.awt.geom.Rectangle2D ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;
import java.util.Date ;

import com.ijchart.xychart.axis.category.CategoryAxis ;
import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.GranttDrawType ;
import com.ijchart.xychart.data.GranttChartData ;
import com.ijchart.xychart.data.IChartData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.shape.GranttChartShape ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.SerialUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ͼ.</zh>
 * <en>A render for simple Gantt charts.</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class GranttRender extends AbstractCategoryRender implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -56526401736677787L ;

  /**
   * <zh>ȱʡĸͼ.</zh>
   * <en>The default draw type of grantt chart.</en>
   */
  public static final GranttDrawType DEFAULT_DRAW_TYPE = GranttDrawType.PARALLELA ;

  /**
   * <zh>ɲֵPaint.</zh>
   * <en>The paint for displaying the percentage complete.</en>
   */
  public static final Paint DEFAULT_COMPLETE_PAINT = Color.GREEN ;

  /**
   * <zh>δɲֵPaint.</zh>
   * <en>The paint for displaying the incomplete part of a task.</en>
   */
  public static final Paint DEFAULT_INCOMPLETE_PAINT = Color.RED ;

  /**
   * <zh>ȱʡɽȵĿʼ߽(ȫĿȰٷֱ).</zh>
   * <en>Controls the starting edge of the progress indicator (expressed as a
   * percentage of the overall bar width).This is a default value.</en>
   */
  public static final double DEFAULT_START_PERCENT = 0.35 ;

  /**
   * <zh>ȱʡɽȵĽ߽(ȫĿȰٷֱ).</zh>
   * <en>Controls the ending edge of the progress indicator (expressed as a
   * percentage of the overall bar width).This is a default value.</en>
   */
  public static final double DEFAULT_END_PERCENT = 0.65 ;

  /**
   * <zh>ȱʡڿǷɽȺδɽ.</zh>
   * <en>A default boolean flag that controls whether or not draw complete
   * and incomplete progress. </en>
   */
  public static final boolean DEFAULT_DRAW_COMPLETE_AND_INCOMPLETE = true ;

  /**
   * <zh>ͼĻ.</zh>
   * <en>The draw type of grantt chart.</en>
   */
  private GranttDrawType drawType ;

  /**
   * <zh>ɲֵPaint.</zh>
   * <en>The paint for displaying the percentage complete.</en>
   */
  private transient Paint completePaint ;

  /**
   * <zh>δɲֵPaint.</zh>
   * <en>The paint for displaying the incomplete part of a task.</en>
   */
  private transient Paint incompletePaint ;

  /**
   * <zh>ɽȵĿʼ߽(ȫĿȰٷֱ).</zh>
   * <en>Controls the starting edge of the progress indicator (expressed as a
   * percentage of the overall bar width).</en>
   */
  private double startPercent ;

  /**
   * <zh>ȱʡɽȵĽ߽(ȫĿȰٷֱ).</zh>
   * <en>Controls the ending edge of the progress indicator (expressed as a
   * percentage of the overall bar width).</en>
   */
  private double endPercent ;

  /**
   * <zh>ڿǷɽȺδɽ.</zh>
   * <en>A boolean flag that controls whether or not draw complete and incomplete progress. </en>
   */
  private boolean drawCompleteAndIncomplete ;

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  public GranttRender(IMatrixDataset dataset) {
    super(dataset) ;

    this.drawType = DEFAULT_DRAW_TYPE ;
    this.completePaint = DEFAULT_COMPLETE_PAINT ;
    this.incompletePaint = DEFAULT_INCOMPLETE_PAINT ;
    this.startPercent = DEFAULT_START_PERCENT ;
    this.endPercent = DEFAULT_END_PERCENT ;
    this.drawCompleteAndIncomplete = DEFAULT_DRAW_COMPLETE_AND_INCOMPLETE ;
  }

  /**
   * <zh>ظͼĻ,ֵΪ{@link com.ijchart.xychart.constant.GranttDrawType#PARALLELA}
   * {@link com.ijchart.xychart.constant.GranttDrawType#VERTICAL}.</zh>
   * <en>Returns the draw type of grantt chart,the type defined in class GranttDrawType.</en>
   *
   * @return GranttDrawType <zh>ͼĻ (Ϊnull).</zh>
   * <en>The draw type of grantt chart (never null).</en>
   */
  public GranttDrawType getDrawType() {
    return this.drawType ;
  }

  /**
   * <zh>øͼĻ.</zh>
   * <en>Sets the draw type of grantt chart.</en>
   *
   * @param drawType GranttDrawType <zh> (Ϊnull).</zh>
   * <en>The draw type (null not permitted).</en>
   */
  public void setDrawType(GranttDrawType drawType) {
    if (drawType == null) {
      throw new IllegalArgumentException("Null 'drawType' argument.") ;
    }

    if (!EqualsUtil.equals(this.drawType, drawType)) {
      this.drawType = drawType ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ػɲֵPaint.</zh>
   * <en>Returns the paint for displaying the percentage complete.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getCompletePaint() {
    return this.completePaint ;
  }

  /**
   * <zh>ûɲֵPaint,
   * м{@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the paint used to show the percentage complete and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setCompletePaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.completePaint = paint ;
    fireRenderChanged() ;
  }

  /**
   * <zh>ػδɲֵPaint.</zh>
   * <en>Returns the paint for displaying the percentage incomplete.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getIncompletePaint() {
    return this.incompletePaint ;
  }

  /**
   * <zh>ûδɲֵPaint,
   * м{@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the paint used to show the percentage incomplete and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setIncompletePaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.incompletePaint = paint ;
    fireRenderChanged() ;
  }

  /**
   * <zh>ؿƽָʾĿʼ߽(ȫĿȰٷֱ).</zh>
   * <en>Returns the position of the start of the progress indicator, as a
   * percentage of the bar width.</en>
   *
   * @return double <zh>ٷֱ.</zh>
   * <en>The start percent.</en>
   */
  public double getStartPercent() {
    return this.startPercent ;
  }

  /**
   * <zh>ÿƽָʾĿʼ߽(ȫĿȰٷֱ),
   * м{@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the position of the start of the progress indicator, as a
   * percentage of the bar width,and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param percent double <zh>ٷֱ.</zh>
   * <en>The percent.</en>
   */
  public void setStartPercent(double percent) {
    if (!(percent >= 0.0 && percent <= 1.0)) {
      throw new IllegalArgumentException("Argument 'percent' must be in [0.0,1.0].") ;
    }

    if (this.startPercent != percent) {
      this.startPercent = percent ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ؿƽָʾĽ߽(ȫĿȰٷֱ).</zh>
   * <en>Returns the position of the end of the progress indicator, as a
   * percentage of the bar width.</en>
   *
   * @return double <zh>ٷֱ.</zh>
   * <en>The end percent.</en>
   */
  public double getEndPercent() {
    return this.endPercent ;
  }

  /**
   * <zh>ÿƽָʾĽ߽(ȫĿȰٷֱ),
   * м{@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the position of the end of the progress indicator, as a percentage
   * of the bar width,and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param percent double <zh>ٷֱ.</zh>
   * <en>The percent.</en>
   */
  public void setEndPercent(double percent) {
    if (!(percent >= 0.0 && percent <= 1.0)) {
      throw new IllegalArgumentException("Argument 'percent' must be in [0.0,1.0].") ;
    }

    if (this.endPercent != percent) {
      this.endPercent = percent ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ڿǷɽȺδɽ.</zh>
   * <en>Returns the boolean flag that controls whether or not draw complete
   * and incomplete progress. </en>
   *
   * @return boolean
   */
  public boolean isDrawCompleteAndIncomplete() {
    return this.drawCompleteAndIncomplete ;
  }

  /**
   * <zh>ڿǷɽȺδɽ.</zh>
   * <en>Sets a boolean flag that controls whether or not draw complete and
   * incomplete progress. </en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setDrawCompleteAndIncomplete(boolean flag) {
    if (this.drawCompleteAndIncomplete != flag) {
      this.drawCompleteAndIncomplete = flag ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>еͼ.</zh>
   * <en>Draws all data items.</en>
   *
   * @param categoryAxis CategoryAxis <zh>category (Ϊnull).</zh>
   * <en>The category axis (null not permitted).</en>
   * @param valueAxis AbstractValueAxis <zh>value (Ϊnull).</zh>
   * <en>The value axis (null not permitted).</en>
   * @param valueAxisEdge RectangleEdge <zh>valueλ (Ϊnull).</zh>
   * <en>The value axis edge (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void drawAllItems(CategoryAxis categoryAxis, AbstractValueAxis valueAxis,
                           RectangleEdge valueAxisEdge, Rectangle2D dataArea,
                           ChartOrientation orientation, RenderShapeCollection renderShapes,
                           float alpha) {
    if (categoryAxis == null) {
      throw new IllegalArgumentException("Null 'categoryAxis' argument.") ;
    }
    if (valueAxis == null) {
      throw new IllegalArgumentException("Null 'valueAxis' argument.") ;
    }
    if (valueAxisEdge == null) {
      throw new IllegalArgumentException("Null 'valueAxisEdge' argument.") ;
    }
    if (dataArea == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }
    if (renderShapes == null) {
      throw new IllegalArgumentException("Null 'renderShapes' argument.") ;
    }
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException("The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    IMatrixDataset dataset = getDataset() ;
    if (dataset.isEmpty()) {
      return ;
    }

    int columns = dataset.getColumnCount() ;
    int rows = dataset.getRowCount() ;
    this.fillPaints = getFillPaints(columns) ;
    getBorderPaints(columns) ;
    getBorderStrokes(columns) ;

    //״ͼ
    if (orientation.equals(ChartOrientation.VERTICAL)) { //ֱ
      double gapSize = 0 ; //С
      if (rows > 1) {
        gapSize = dataArea.getWidth() * categoryAxis.getCategoryMargin() / (rows - 1) ;
      }
      double axisMargin = categoryAxis.getLowerMargin() + categoryAxis.getUpperMargin()
                          + ((rows == 1) ? 0 : categoryAxis.getCategoryMargin()) ; //еĿհ
      double categoryWidth = dataArea.getWidth() * (1 - axisMargin) / rows ; //ÿϵеĿ
      double categoryOneMargin = (dataArea.getWidth()
                                  * (1 - categoryAxis.getLowerMargin()
                                     - categoryAxis.getUpperMargin())
                                  * getItemMargin()) / rows ; //ÿcategoryĿհ

      double calculateBarWidth = 0 ; //Ŀ
      double barItemMargin = 0 ; //ÿ֮Ŀ
      if (drawType.equals(GranttDrawType.PARALLELA)) { //ˮƽ
        calculateBarWidth = (categoryWidth - categoryOneMargin) / columns ;
        barItemMargin = categoryOneMargin / columns ;
      } else if (drawType.equals(GranttDrawType.VERTICAL)) {
        calculateBarWidth = categoryWidth ;
        barItemMargin = 0 ;
      }
      double reallyBarWidth = Math.min(calculateBarWidth, getMaximumBarWidth()) ; //ʵʿ
      if (reallyBarWidth < getMinimumBarWidth()) { //ĿССƿ
        return ;
      }

      for (int i = 0 ; i < rows ; i++) {
        double startX = dataArea.getX() + dataArea.getWidth() * categoryAxis.getLowerMargin()
                        + i * (categoryWidth + gapSize) ;
        int currRowCols = dataset.getColumnCount(i) ;

        for (int j = 0 ; j < currRowCols ; j++) {
          IChartData data = dataset.getData(i, j) ;
          if (data != null && data instanceof GranttChartData) {
            GranttChartData granttData = (GranttChartData) data ;
            Date startDate = granttData.getStart() ;
            Date endDate = granttData.getEnd() ;
            if (startDate == null || endDate == null) {
              renderShapes.addDataItemShape(null) ;
            } else {
              long startMills = startDate.getTime() ;
              long endMills = endDate.getTime() ;
              double startPos = valueAxis.valueToJava2D(dataArea, valueAxisEdge, startMills, false) ;
              double endPos = valueAxis.valueToJava2D(dataArea, valueAxisEdge, endMills, false) ;
              //ֵ᷶Χֵ
              if (startPos == AbstractValueAxis.INVALID_COORDINATE_VALUE
                  || endPos == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
                renderShapes.addDataItemShape(null) ;
                continue ;
              }

              double x11 = 0 ;
              if (drawType.equals(GranttDrawType.PARALLELA)) { //ˮƽ
                x11 = startX + 0.5 * barItemMargin + 0.5 * (calculateBarWidth - reallyBarWidth)
                      + (calculateBarWidth + barItemMargin) * j ;
              } else if (drawType.equals(GranttDrawType.VERTICAL)) {
                x11 = startX + 0.5 * barItemMargin + 0.5 * (calculateBarWidth - reallyBarWidth) ;
              }

              Rectangle2D granttRect = new Rectangle2D.Double(x11, Math.min(startPos, endPos),
                                                              reallyBarWidth,
                                                              Math.abs(endPos - startPos)) ;
              Rectangle2D completeRect = getCompleteRect(granttRect, granttData.getPercentComplete(),
                                                         orientation, valueAxis.isInverted()) ;
              Rectangle2D incompleteRect = getIncompleteRect(granttRect,
                                                             granttData.getPercentComplete(),
                                                             orientation, valueAxis.isInverted()) ;

              GranttChartShape granttShape = new GranttChartShape(granttRect, completeRect,
                                                                  incompleteRect) ;
              granttShape.setCompleteAreaPaint(this.completePaint) ;
              granttShape.setIncompleteAreaPaint(this.incompletePaint) ;
              IChartShape shape = createItemShape(orientation, valueAxis, i, j, granttRect,
                                                  fillPaints[j], null, null, alpha, granttShape) ;
              renderShapes.addDataItemShape(shape) ;
            }
          } else {
            renderShapes.addDataItemShape(null) ;
          }
        }
      }

    } else if (orientation.equals(ChartOrientation.HORIZONTAL)) { //ˮƽ
      double gapSize = 0 ; //С
      if (rows > 1) {
        gapSize = dataArea.getHeight() * categoryAxis.getCategoryMargin() / (rows - 1) ;
      }
      double axisMargin = categoryAxis.getLowerMargin() + categoryAxis.getUpperMargin()
                          + ((rows == 1) ? 0 : categoryAxis.getCategoryMargin()) ; //еĿհ
      double categoryWidth = dataArea.getHeight() * (1 - axisMargin) / rows ; //ÿϵеĿ
      double categoryOneMargin = (dataArea.getHeight()
                                  * (1 - categoryAxis.getLowerMargin()
                                     - categoryAxis.getUpperMargin())
                                  * getItemMargin()) / rows ; //ÿcategoryĿհ

      double calculateBarWidth = 0 ; //Ŀ
      double barItemMargin = 0 ; //ÿ֮Ŀ
      if (drawType.equals(GranttDrawType.PARALLELA)) { //ˮƽ
        calculateBarWidth = (categoryWidth - categoryOneMargin) / columns ;
        barItemMargin = categoryOneMargin / columns ;
      } else if (drawType.equals(GranttDrawType.VERTICAL)) {
        calculateBarWidth = categoryWidth ;
        barItemMargin = 0 ;
      }

      double reallyBarWidth = Math.min(calculateBarWidth, getMaximumBarWidth()) ; //ʵʿ
      if (reallyBarWidth < getMinimumBarWidth()) { //ĿССƿ
        return ;
      }

      for (int i = 0 ; i < rows ; i++) {
        double startY = dataArea.getY() + dataArea.getHeight() * categoryAxis.getUpperMargin()
                        + i * (categoryWidth + gapSize) ;
        int currRowCols = dataset.getColumnCount(i) ;

        for (int j = 0 ; j < currRowCols ; j++) {
          IChartData data = dataset.getData(i, j) ;
          if (data != null && data instanceof GranttChartData) {
            GranttChartData granttData = (GranttChartData) data ;
            Date startDate = granttData.getStart() ;
            Date endDate = granttData.getEnd() ;
            if (startDate == null || endDate == null) {
              renderShapes.addDataItemShape(null) ;
            } else {
              long startMills = startDate.getTime() ;
              long endMills = endDate.getTime() ;
              double startPos = valueAxis.valueToJava2D(dataArea, valueAxisEdge, startMills, false) ;
              double endPos = valueAxis.valueToJava2D(dataArea, valueAxisEdge, endMills, false) ;
              //ֵ᷶Χֵ
              if (startPos == AbstractValueAxis.INVALID_COORDINATE_VALUE
                  || endPos == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
                renderShapes.addDataItemShape(null) ;
                continue ;
              }

              double y11 = 0 ;
              if (drawType.equals(GranttDrawType.PARALLELA)) { //ˮƽ
                y11 = startY + 0.5 * barItemMargin + 0.5 * (calculateBarWidth - reallyBarWidth)
                      + (calculateBarWidth + barItemMargin) * j ;
              } else {
                y11 = startY + 0.5 * barItemMargin + 0.5 * (calculateBarWidth - reallyBarWidth) ;
              }
              Rectangle2D granttRect = new Rectangle2D.Double(Math.min(startPos, endPos), y11,
                                                              Math.abs(endPos - startPos),
                                                              reallyBarWidth) ;
              Rectangle2D completeRect = getCompleteRect(granttRect, granttData.getPercentComplete(),
                                                         orientation, valueAxis.isInverted()) ;
              Rectangle2D incompleteRect = getIncompleteRect(granttRect,
                                                             granttData.getPercentComplete(),
                                                             orientation, valueAxis.isInverted()) ;

              GranttChartShape granttShape = new GranttChartShape(granttRect, completeRect,
                                                                  incompleteRect) ;
              granttShape.setCompleteAreaPaint(this.completePaint) ;
              granttShape.setIncompleteAreaPaint(this.incompletePaint) ;

              IChartShape shape = createItemShape(orientation, valueAxis, i, j, granttRect,
                                                  fillPaints[j], null, null, alpha, granttShape) ;
              renderShapes.addDataItemShape(shape) ;
            }
          } else {
            renderShapes.addDataItemShape(null) ;
          }
        }
      }
    }

  }

  /**
   * <zh>ɲֵľ.</zh>
   * <en>Calculate the rectangle of complete.</en>
   *
   * @param granttRect Rectangle2D
   * @param completePercent Double
   * @param orientation ChartOrientation
   * @param invert boolean
   * @return Rectangle2D
   */
  private Rectangle2D getCompleteRect(Rectangle2D granttRect, Double completePercent,
                                      ChartOrientation orientation, boolean invert) {
    Rectangle2D result = null ;

    if (granttRect != null && completePercent != null && orientation != null) {
      double complete = completePercent.doubleValue() ;
      double minStartPercent = Math.min(this.startPercent, this.endPercent) ;
      if (orientation.equals(ChartOrientation.VERTICAL)) { //ֱ
        double x = granttRect.getX() + granttRect.getWidth() * minStartPercent ;
        double w = granttRect.getWidth() * Math.abs(this.endPercent - this.startPercent) ;
        double h = granttRect.getHeight() * complete ;
        if (invert) {
          double y = granttRect.getY() ;
          result = new Rectangle2D.Double(x, y, w, h) ;
        } else {
          double y = granttRect.getY() + (1 - complete) * granttRect.getHeight() ;
          result = new Rectangle2D.Double(x, y, w, h) ;
        }
      } else if (orientation.equals(ChartOrientation.HORIZONTAL)) {
        double y = granttRect.getY() + granttRect.getHeight() * minStartPercent ;
        double w = granttRect.getWidth() * complete ;
        double h = granttRect.getHeight() * Math.abs(this.endPercent - this.startPercent) ;
        if (invert) {
          double x = granttRect.getX() + granttRect.getWidth() * (1 - complete) ;
          result = new Rectangle2D.Double(x, y, w, h) ;
        } else {
          double x = granttRect.getX() ;
          result = new Rectangle2D.Double(x, y, w, h) ;
        }
      }
    }

    return result ;
  }

  /**
   * <zh>δɲֵľ.</zh>
   * <en>Calculate the rectangle of incomplete.</en>
   *
   * @param granttRect Rectangle2D
   * @param completePercent Double
   * @param orientation ChartOrientation
   * @param invert boolean
   * @return Rectangle2D
   */
  private Rectangle2D getIncompleteRect(Rectangle2D granttRect, Double completePercent,
                                        ChartOrientation orientation, boolean invert) {
    Rectangle2D result = null ;

    if (granttRect != null && completePercent != null && orientation != null) {
      double complete = completePercent.doubleValue() ;
      double minStartPercent = Math.min(this.startPercent, this.endPercent) ;
      if (orientation.equals(ChartOrientation.VERTICAL)) { //ֱ
        double x = granttRect.getX() + granttRect.getWidth() * minStartPercent ;
        double w = granttRect.getWidth() * Math.abs(this.endPercent - this.startPercent) ;
        double h = granttRect.getHeight() * (1 - complete) ;
        if (invert) {
          double y = granttRect.getY() + complete * granttRect.getHeight() ;
          result = new Rectangle2D.Double(x, y, w, h) ;
        } else {
          double y = granttRect.getY() ;
          result = new Rectangle2D.Double(x, y, w, h) ;
        }
      } else if (orientation.equals(ChartOrientation.HORIZONTAL)) {
        double w = granttRect.getWidth() * (1 - complete) ;
        double h = granttRect.getHeight() * Math.abs(this.endPercent - this.startPercent) ;
        double y = granttRect.getY() + granttRect.getHeight() * minStartPercent ;
        if (invert) {
          double x = granttRect.getX() ;
          result = new Rectangle2D.Double(x, y, w, h) ;
        } else {
          double x = granttRect.getX() + granttRect.getWidth() * complete ;
          result = new Rectangle2D.Double(x, y, w, h) ;
        }
      }
    }

    return result ;
  }

  /**
   * <zh>ֻͼ,ͼε.</zh>
   * <en>Create a shape with no attribute specified.</en>
   *
   * @param rect Rectangle2D <zh>ͼľ (Ϊnull).</zh>
   * <en>The bar rectangle (null not permitted).</en>
   * @param text String <zh>Ҫʾ (Ϊnull).</zh>
   * <en>The text (null permitted).</en>
   * @param appendParam Object <zh>ӵĲ,ڴݶĶ (Ϊnull).</zh>
   * <en>The append parameter,used for special propose (null permitted).</en>
   *
   * @return IChartShape <zh>ͼ.</zh>
   * <en>The shape.</en>
   */
  protected IChartShape createItemShapeOnly(Rectangle2D rect, String text,
                                            Object appendParam) {
    if (appendParam == null) {
      throw new IllegalArgumentException("Null 'appendParam' argument.") ;
    }

    return (IChartShape) appendParam ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof GranttRender)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    GranttRender that = (GranttRender) obj ;
    if (EqualsUtil.equals(this.drawType, that.drawType)) {
      return false ;
    }
    if (EqualsUtil.equals(this.completePaint, that.completePaint)) {
      return false ;
    }
    if (EqualsUtil.equals(this.incompletePaint, that.incompletePaint)) {
      return false ;
    }
    if (this.startPercent != that.startPercent) {
      return false ;
    }
    if (this.endPercent != that.endPercent) {
      return false ;
    }
    if (this.drawCompleteAndIncomplete != that.drawCompleteAndIncomplete) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double result = super.hashCode() ;
    result = 37 * result + HashcodeUtil.hashCode(this.drawType) ;
    result = 37 * result + HashcodeUtil.hashCode(this.completePaint) ;
    result = 37 * result + HashcodeUtil.hashCode(this.incompletePaint) ;
    result = 37 * result + 1000 * this.startPercent ;
    result = 37 * result + 1000 * this.endPercent ;
    result = 37 * result + HashcodeUtil.hashCode(this.drawCompleteAndIncomplete) ;

    return (int) result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[drawType=" + this.drawType
      + ",completePaint=" + this.completePaint
      + ",incompletePaint=" + this.incompletePaint
      + ",startPercent=" + this.startPercent
      + ",endPercent=" + this.endPercent
      + ",drawCompleteAndIncomplete=" + this.drawCompleteAndIncomplete
      + "]" ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the output stream.
   *
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;

    SerialUtil.writePaint(stream, this.completePaint) ;
    SerialUtil.writePaint(stream, this.incompletePaint) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;

    this.completePaint = SerialUtil.readPaint(stream) ;
    this.incompletePaint = SerialUtil.readPaint(stream) ;
  }
}
