package com.ijchart.xychart.render.category ;

import java.awt.Color ;
import java.awt.GradientPaint ;
import java.awt.Paint ;
import java.awt.geom.Rectangle2D ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;
import java.util.ArrayList ;
import java.util.List ;

import com.ijchart.xychart.axis.category.CategoryAxis ;
import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.GradientPaintTransformType ;
import com.ijchart.xychart.constant.LegendType ;
import com.ijchart.xychart.data.IChartData ;
import com.ijchart.xychart.data.NumberChartData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.SerialUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ٲ͵״ͼ.</zh>
 * <en>Draw waterfall bar shapes.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class WaterfallBarRender extends AbstractCategoryRender implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -8053211674538308465L ;

  /**
   * <zh>ȱʡƵһ״ͼPaint.</zh>
   * <en>The default paint used to draw the first bar. </en>
   */
  public static final Paint DEFAULT_FIRST_BAR_PAINT = new GradientPaint
                                                      (0.0f, 0.0f, new Color(0x22, 0x22, 0xFF),
                                                       0.0f, 0.0f, new Color(0x66, 0x66, 0xFF)) ;

  /**
   * <zh>ȱʡһ״ͼPaint.</zh>
   * <en>The default paint used to draw the last bar.</en>
   */
  public static final Paint DEFAULT_LAST_BAR_PAINT = new GradientPaint
                                                     (0.0f, 0.0f, new Color(0xFF, 0xFF, 0x22),
                                                      0.0f, 0.0f, new Color(0xFF, 0xFF, 0x66)) ;

  /**
   * <zh>ȱʡֵ״ͼPaint.</zh>
   * <en>The default paint used to draw bars having positive values.</en>
   */
  public static final Paint DEFAULT_POSITIVE_BAR_PAINT = new GradientPaint
                                                         (0.0f, 0.0f, new Color(0x22, 0xFF, 0x22),
                                                          0.0f, 0.0f, new Color(0x66, 0xFF, 0x66)) ;

  /**
   * <zh>ȱʡƸֵ״ͼPaint.</zh>
   * <en>The default paint used to draw bars having negative values.</en>
   */
  public static final Paint DEFAULT_NEGATIVE_BAR_PAINT = new GradientPaint
                                                         (0.0f, 0.0f, new Color(0xFF, 0x22, 0x22),
                                                          0.0f, 0.0f, new Color(0xFF, 0x66, 0x66)) ;

  /**
   * <zh>Ƶһ״ͼPaint.</zh>
   * <en>The paint used to draw the first bar. </en>
   */
  private transient Paint firstBarPaint ;

  /**
   * <zh>һ״ͼPaint.</zh>
   * <en>The paint used to draw the last bar.</en>
   */
  private transient Paint lastBarPaint ;

  /**
   * <zh>ֵ״ͼPaint.</zh>
   * <en>The paint used to draw bars having positive values.</en>
   */
  private transient Paint positiveBarPaint ;

  /**
   * <zh>Ƹֵ״ͼPaint.</zh>
   * <en>The paint used to draw bars having negative values.</en>
   */
  private transient Paint negativeBarPaint ;

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  public WaterfallBarRender(IMatrixDataset dataset) {
    this(dataset, DEFAULT_FIRST_BAR_PAINT, DEFAULT_POSITIVE_BAR_PAINT,
         DEFAULT_NEGATIVE_BAR_PAINT, DEFAULT_LAST_BAR_PAINT) ;
  }

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   * @param firstBarPaint Paint <zh>Ƶһ״ͼPaint.</zh>
   * <en>The paint used to draw the first bar. </en>
   * @param positiveBarPaint Paint <zh>һ״ͼPaint.</zh>
   * <en>The paint used to draw the last bar.</en>
   * @param negativeBarPaint Paint <zh>ֵ״ͼPaint.</zh>
   * <en>The paint used to draw bars having positive values.</en>
   * @param lastBarPaint Paint <zh>Ƹֵ״ͼPaint.</zh>
   * <en>The paint used to draw bars having negative values.</en>
   */
  public WaterfallBarRender(IMatrixDataset dataset, Paint firstBarPaint, Paint positiveBarPaint,
                            Paint negativeBarPaint, Paint lastBarPaint) {
    super(dataset) ;

    if (firstBarPaint == null) {
      throw new IllegalArgumentException("Null 'firstBarPaint' argument") ;
    }
    if (positiveBarPaint == null) {
      throw new IllegalArgumentException("Null 'positiveBarPaint' argument") ;
    }
    if (negativeBarPaint == null) {
      throw new IllegalArgumentException("Null 'negativeBarPaint' argument") ;
    }
    if (lastBarPaint == null) {
      throw new IllegalArgumentException("Null 'lastBarPaint' argument") ;
    }

    this.firstBarPaint = firstBarPaint ;
    this.lastBarPaint = lastBarPaint ;
    this.positiveBarPaint = positiveBarPaint ;
    this.negativeBarPaint = negativeBarPaint ;

    setBaseItemFillPaintTransformer(GradientPaintTransformType.CENTER_VERTICAL) ;
    setMinimumBarWidth(1.0) ;
  }

  /**
   * <zh>ػƵһ״ͼPaint.</zh>
   * <en>Returns the paint used to draw the first bar.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getFirstBarPaint() {
    return this.firstBarPaint ;
  }

  /**
   * <zh>ûƵһ״ͼPaint,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the paint that will be used to draw the first bar and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setFirstBarPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument") ;
    }

    this.firstBarPaint = paint ;
    fireRenderChanged() ;
  }

  /**
   * <zh>ػһ״ͼPaint.</zh>
   * <en>Returns the paint used to draw the last bar.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getLastBarPaint() {
    return this.lastBarPaint ;
  }

  /**
   * <zh>ûһ״ͼPaint,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the paint that will be used to draw the last bar and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setLastBarPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument") ;
    }
    this.lastBarPaint = paint ;

    fireRenderChanged() ;
  }

  /**
   * <zh>ֵΪ״ͼPaint.</zh>
   * <en>Returns the paint used to draw bars with positive values.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getPositiveBarPaint() {
    return this.positiveBarPaint ;
  }

  /**
   * <zh>ûֵΪ״ͼPaint,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the paint that will be used to draw bars with positive values and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setPositiveBarPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument") ;
    }

    this.positiveBarPaint = paint ;
    fireRenderChanged() ;
  }

  /**
   * <zh>ֵΪ״ͼPaint.</zh>
   * <en>Returns the paint used to draw bars with negative values.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getNegativeBarPaint() {
    return this.negativeBarPaint ;
  }

  /**
   * <zh>ûֵΪ״ͼPaint,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the paint that will be used to draw bars with positive values and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setNegativeBarPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument") ;
    }

    this.negativeBarPaint = paint ;
    fireRenderChanged() ;
  }

  /**
   * <zh>еͼ.</zh>
   * <en>Draws all data items.</en>
   *
   * @param categoryAxis CategoryAxis <zh>category (Ϊnull).</zh>
   * <en>The category axis (null not permitted).</en>
   * @param valueAxis AbstractValueAxis <zh>value (Ϊnull).</zh>
   * <en>The value axis (null not permitted).</en>
   * @param valueAxisEdge RectangleEdge <zh>valueλ (Ϊnull).</zh>
   * <en>The value axis edge (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void drawAllItems(CategoryAxis categoryAxis, AbstractValueAxis valueAxis,
                           RectangleEdge valueAxisEdge, Rectangle2D dataArea,
                           ChartOrientation orientation, RenderShapeCollection renderShapes,
                           float alpha) {
    if (categoryAxis == null) {
      throw new IllegalArgumentException("Null 'categoryAxis' argument.") ;
    }
    if (valueAxis == null) {
      throw new IllegalArgumentException("Null 'valueAxis' argument.") ;
    }
    if (valueAxisEdge == null) {
      throw new IllegalArgumentException("Null 'valueAxisEdge' argument.") ;
    }
    if (dataArea == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }
    if (renderShapes == null) {
      throw new IllegalArgumentException("Null 'renderShapes' argument.") ;
    }
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException("The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    IMatrixDataset dataset = getDataset() ;
    if (dataset.isEmpty()) {
      return ;
    }
    int columns = dataset.getColumnCount() ;
    int rows = dataset.getRowCount() ;

    double basePosition = valueAxis.valueToJava2D(dataArea, valueAxisEdge, 0, false) ;
    if (basePosition == AbstractValueAxis.INVALID_COORDINATE_VALUE) { //ֵ᷶Χ
      return ;
    }

    this.fillPaints = getFillPaints(columns) ;
    this.borderPaints = getBorderPaints(columns) ;
    this.borderStrokes = getBorderStrokes(columns) ;

    //״ͼ
    if (orientation.equals(ChartOrientation.VERTICAL)) { //ֱ
      double gapSize = 0 ; //С
      if (rows > 1) {
        gapSize = dataArea.getWidth() * categoryAxis.getCategoryMargin() / (rows - 1) ;
      }
      double axisMargin = categoryAxis.getLowerMargin() + categoryAxis.getUpperMargin()
                          + ((rows == 1) ? 0 : categoryAxis.getCategoryMargin()) ; //еĿհ
      double categoryWidth = dataArea.getWidth() * (1 - axisMargin) / rows ; //ÿϵеĿ

      double calculateBarWidth = categoryWidth / columns ; //Ŀ
      double reallyBarWidth = Math.min(calculateBarWidth, getMaximumBarWidth()) ; //ʵʿ
      if (reallyBarWidth < getMinimumBarWidth()) { //ĿССƿ
        return ;
      }

      for (int i = 0 ; i < rows ; i++) {
        double startX = dataArea.getX() + dataArea.getWidth() * categoryAxis.getLowerMargin()
                        + i * (categoryWidth + gapSize) ;

        double currentYPos = basePosition ; //ǰYλ
        int currRowCols = dataset.getColumnCount(i) ;

        for (int j = 0 ; j < currRowCols ; j++) {
          IChartData data = dataset.getData(i, j) ;
          if (data != null && data instanceof NumberChartData) {
            double value = ((NumberChartData) data).getValue() ;
            boolean firstOrLast = (j == 0 || j == columns - 1) ;
            double y = valueAxis.valueToJava2D(dataArea, valueAxisEdge, value, firstOrLast) ;
            //ֵ᷶Χֵ
            if (y == AbstractValueAxis.INVALID_COORDINATE_VALUE || y == basePosition) {
              renderShapes.addDataItemShape(null) ;
              continue ;
            }

            double x11 = startX + 0.5 * (calculateBarWidth - reallyBarWidth)
                         + calculateBarWidth * j ;

            double rectY = 0 ;
            if (j == 0) { //һֵ
              currentYPos = y ;
              rectY = Math.min(y, basePosition) ;
            } else if (j == columns - 1) { //һֵ
              rectY = Math.min(y, basePosition) ;
            } else { //ǵһһֵ
              if (y <= basePosition) {
                rectY = currentYPos - Math.abs(y - basePosition) ;
                currentYPos -= Math.abs(basePosition - y) ;
              } else {
                rectY = currentYPos ;
                currentYPos += Math.abs(basePosition - y) ;
              }
            }

            Rectangle2D rect = new Rectangle2D.Double(x11, rectY, reallyBarWidth,
                                                      Math.abs(basePosition - y)) ;
            IChartShape shape = createItemShape(orientation, valueAxis, i, j, rect, fillPaints[j],
                                                borderPaints[j], borderStrokes[j], alpha, null) ;
            renderShapes.addDataItemShape(shape) ;
          } else {
            renderShapes.addDataItemShape(null) ;
          }
        }
      }
    } else if (orientation.equals(ChartOrientation.HORIZONTAL)) { //ˮƽ
      double gapSize = 0 ; //С
      if (rows > 1) {
        gapSize = dataArea.getHeight() * categoryAxis.getCategoryMargin() / (rows - 1) ;
      }
      double axisMargin = categoryAxis.getLowerMargin() + categoryAxis.getUpperMargin()
                          + ((rows == 1) ? 0 : categoryAxis.getCategoryMargin()) ; //еĿհ
      double categoryWidth = dataArea.getHeight() * (1 - axisMargin) / rows ; //ÿϵеĿ

      double calculateBarWidth = categoryWidth / columns ; //Ŀ
      double reallyBarWidth = Math.min(calculateBarWidth, getMaximumBarWidth()) ; //ʵʿ
      if (reallyBarWidth < getMinimumBarWidth()) { //ĿССƿ
        return ;
      }

      for (int i = 0 ; i < rows ; i++) {
        double startY = dataArea.getY() + dataArea.getHeight() * categoryAxis.getUpperMargin()
                        + i * (categoryWidth + gapSize) ;

        double currentYPos = basePosition ; //ǰYλ
        int currRowCols = dataset.getColumnCount(i) ;

        for (int j = 0 ; j < currRowCols ; j++) {
          IChartData data = dataset.getData(i, j) ;
          if (data != null && data instanceof NumberChartData) {
            double value = ((NumberChartData) data).getValue() ;
            boolean firstOrLast = (j == 0 || j == columns - 1) ;
            double x = valueAxis.valueToJava2D(dataArea, valueAxisEdge, value, firstOrLast) ;
            //ֵ᷶Χֵ
            if (x == AbstractValueAxis.INVALID_COORDINATE_VALUE || x == basePosition) {
              renderShapes.addDataItemShape(null) ;
              continue ;
            }

            double y11 = startY + +0.5 * (calculateBarWidth - reallyBarWidth)
                         + calculateBarWidth * j ;

            double rectX = 0 ;
            if (j == 0) { //һֵ
              currentYPos = x ;
              rectX = Math.min(x, basePosition) ;
            } else if (j == columns - 1) { //һֵ
              rectX = Math.min(x, basePosition) ;
            } else { //ǵһһֵ
              if (x <= basePosition) {
                rectX = currentYPos - Math.abs(x - basePosition) ;
                currentYPos -= Math.abs(basePosition - x) ;
              } else {
                rectX = currentYPos ;
                currentYPos += Math.abs(basePosition - x) ;
              }
            }

            Rectangle2D rect = new Rectangle2D.Double(rectX, y11,
                                                      Math.abs(basePosition - x), reallyBarWidth) ;
            IChartShape shape = createItemShape(orientation, valueAxis, i, j, rect, fillPaints[j],
                                                borderPaints[j], borderStrokes[j], alpha, null) ;
            renderShapes.addDataItemShape(shape) ;
          } else {
            renderShapes.addDataItemShape(null) ;
          }
        }
      }
    }

  }

  /**
   * <zh>õPaintİ.</zh>
   * <en>Helper method for returns the fill paint.</en>
   *
   * @param autoFillPaint Paint <zh>ԶѡPaint.</zh>
   * <en>The auto selected fill paint.</en>
   * @param row int <zh>ݼ.</zh>
   * <en>The dataset row number.</en>
   * @param column int <zh>ݼ.</zh>
   * <en>The dataset column number.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  protected Paint getItemFillPaintHelper(Paint autoFillPaint, int row, int column) {
    Paint result = null ;

    IMatrixDataset dataset = getDataset() ;
    int cols = dataset.getColumnCount() ;
    List list = new ArrayList() ;
    for (int i = 0 ; i < cols ; i++) {
      list.add(dataset.getData(row, i)) ;
    }
    for (int i = cols - 1 ; i >= 0 ; i--) {
      if (list.get(i) == null) {
        list.remove(i) ;
      } else {
        break ;
      }
    }

    if (list != null) {
      if (column >= 0 && column <= list.size() - 1) {
        if (column == 0) {
          result = this.firstBarPaint ;
        } else if (column == list.size() - 1) {
          result = this.lastBarPaint ;
        } else {
          Object object = list.get(column) ;
          if (object != null && object instanceof NumberChartData) {
            NumberChartData value = (NumberChartData) object ;
            double val = value.getValue() ;
            if (val >= 0) {
              result = this.positiveBarPaint ;
            } else {
              result = this.negativeBarPaint ;
            }
          }
        }
      }
    }

    return result ;
  }

  /**
   * <zh>ͼ.</zh>
   * <en>Return the legend type.</en>
   *
   * @return LegendType <zh>ͼ.</zh>
   * <en>The legend type.</en>
   */
  public LegendType getLegendType() {
    return LegendType.ROW ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof WaterfallBarRender)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    WaterfallBarRender that = (WaterfallBarRender) obj ;
    if (!EqualsUtil.equals(this.firstBarPaint, that.firstBarPaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.lastBarPaint, that.lastBarPaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.positiveBarPaint, that.positiveBarPaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.negativeBarPaint, that.negativeBarPaint)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    int result = super.hashCode() ;
    result = result * 37 + HashcodeUtil.hashCode(this.firstBarPaint) ;
    result = result * 37 + HashcodeUtil.hashCode(this.lastBarPaint) ;
    result = result * 37 + HashcodeUtil.hashCode(this.negativeBarPaint) ;
    result = result * 37 + HashcodeUtil.hashCode(this.positiveBarPaint) ;

    return result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the output stream.
   *
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;
    SerialUtil.writePaint(stream, this.firstBarPaint) ;
    SerialUtil.writePaint(stream, this.lastBarPaint) ;
    SerialUtil.writePaint(stream, this.positiveBarPaint) ;
    SerialUtil.writePaint(stream, this.negativeBarPaint) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;
    this.firstBarPaint = SerialUtil.readPaint(stream) ;
    this.lastBarPaint = SerialUtil.readPaint(stream) ;
    this.positiveBarPaint = SerialUtil.readPaint(stream) ;
    this.negativeBarPaint = SerialUtil.readPaint(stream) ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[firstBarPaint=" + firstBarPaint
      + ",negativeBarPaint=" + negativeBarPaint
      + ",positiveBarPaint=" + positiveBarPaint
      + ",lastBarPaint=" + lastBarPaint + "]" ;
  }
}
