package com.ijchart.xychart.shape ;

import java.awt.AlphaComposite ;
import java.awt.Graphics2D ;
import java.awt.Shape ;
import java.awt.geom.AffineTransform ;
import java.awt.geom.Ellipse2D ;
import java.awt.geom.GeneralPath ;
import java.awt.geom.Point2D ;
import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;

import com.ijchart.xychart.constant.ShapeType ;
import com.ijchart.xychart.ui.Graphics2DState ;
import com.ijchart.xychart.util.AnchorUtil ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ERRORʮͼ.</zh>
 * <en>Draw an error cross {@link com.ijchart.xychart.shape.IChartShape}.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class ErrorCrossChartShape extends AbstractChartShape implements Serializable, Cloneable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -2021724350541778888L ;

  /**
   * <zh>ȱʡǷxError.</zh>
   * <en>A default flag that controls whether or not the x-error bars are drawn.</en>
   */
  public static final boolean DEFAULT_DRAW_X_ERROR = true ;

  /**
   * <zh>ȱʡǷyError.</zh>
   * <en>A default flag that controls whether or not the y-error bars are drawn.</en>
   */
  public static final boolean DEFAULT_DRAW_Y_ERROR = true ;

  /**
   * <zh>ȱʡxyߵյĽߵĳ.</zh>
   * <en>The default length of the cap at the end of the error bars.</en>
   */
  public static final double DEFAULT_CAP_LENGTH = 4 ;

  /**
   * <zh>x.</zh>
   * <en>The coordinate of start x.</en>
   */
  private double startX ;

  /**
   * <zh>xյ.</zh>
   * <en>The coordinate of end x.</en>
   */
  private double endX ;

  /**
   * <zh>ĵx.</zh>
   * <en>The coordinate of center x.</en>
   */
  private double centerX ;

  /**
   * <zh>ĵy.</zh>
   * <en>The coordinate of center y.</en>
   */
  private double centerY ;

  /**
   * <zh>y.</zh>
   * <en>The coordinate of start y.</en>
   */
  private double startY ;

  /**
   * <zh>yյ.</zh>
   * <en>The coordinate of end y.</en>
   */
  private double endY ;

  /**
   * <zh>ĵͼ.</zh>
   * <en>The shape of center point.</en>
   */
  private Shape centerShape ;

  /**
   * <zh>ǷxError.</zh>
   * <en>A flag that controls whether or not the x-error bars are drawn.</en>
   */
  private boolean drawXError ;

  /**
   * <zh>ǷyError.</zh>
   * <en>A flag that controls whether or not the y-error bars are drawn.</en>
   */
  private boolean drawYError ;

  /**
   * <zh>xyߵյĽߵĳ.</zh>
   * <en>The length of the cap at the end of the error bars.</en>
   */
  private double capLength ;

  /**
   * <zh>ͼ.</zh>
   * <en>Constructor.</en>
   *
   * @param startX double <zh>x.</zh>
   * <en>The coordinate of start x.</en>
   * @param endX double <zh>xյ.</zh>
   * <en>The coordinate of end x.</en>
   * @param centerX double <zh>ĵx.</zh>
   * <en>The coordinate of center x.</en>
   * @param centerY double <zh>ĵy.</zh>
   * <en>The coordinate of center y.</en>
   * @param startY double <zh>y.</zh>
   * <en>The coordinate of start y.</en>
   * @param endY double <zh>yյ.</zh>
   * <en>The coordinate of end y.</en>
   * @param centerShape Shape <zh>ĵͼ (Ϊnull).</zh>
   * <en>The shape of center point (null permitted).</en>
   */
  public ErrorCrossChartShape(double startX, double endX, double centerX, double centerY,
                              double startY, double endY, Shape centerShape) {
    this.startX = startX ;
    this.endX = endX ;
    this.centerX = centerX ;
    this.centerY = centerY ;
    this.startY = startY ;
    this.endY = endY ;
    this.centerShape = centerShape ;

    this.drawXError = DEFAULT_DRAW_X_ERROR ;
    this.drawYError = DEFAULT_DRAW_Y_ERROR ;
    this.capLength = DEFAULT_CAP_LENGTH ;

    setDrawBorder(true) ;
    setFillShape(true) ;
  }

  /**
   * <zh>x.</zh>
   * <en>Returns the coordinate of start x.</en>
   *
   * @return double <zh>x.</zh>
   * <en>The start x coordinate.</en>
   */
  public double getStartX() {
    return this.startX ;
  }

  /**
   * <zh>x,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the coordinate of start x,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param startX double <zh>x.</zh>
   * <en>The start x coordinate.</en>
   */
  public void setStartX(double startX) {
    if (this.startX != startX) {
      this.startX = startX ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>xյ.</zh>
   * <en>Returns the coordinate of end x.</en>
   *
   * @return double <zh>xյ.</zh>
   * <en>The end x coordinate.</en>
   */
  public double getEndX() {
    return this.endX ;
  }

  /**
   * <zh>xյ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the coordinate of end x,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param endX double <zh>xյ.</zh>
   * <en>The end x coordinate.</en>
   */
  public void setEndX(double endX) {
    if (this.endX != endX) {
      this.endX = endX ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ĵx.</zh>
   * <en>Returns the coordinate of center x.</en>
   *
   * @return double <zh>x.</zh>
   * <en>The center x coordinate.</en>
   */
  public double getCenterX() {
    return this.centerX ;
  }

  /**
   * <zh>ĵx,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the coordinate of center x,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param centerX double <zh>x.</zh>
   * <en>The center x coordinate.</en>
   */
  public void setCenterX(double centerX) {
    if (this.centerX != centerX) {
      this.centerX = centerX ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ĵy.</zh>
   * <en>Returns the coordinate of center y.</en>
   *
   * @return double <zh>y.</zh>
   * <en>The center y coordinate.</en>
   */
  public double getCenterY() {
    return this.centerY ;
  }

  /**
   * <zh>ĵy,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the coordinate of center y,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param centerY double <zh>y.</zh>
   * <en>The center y coordinate.</en>
   */
  public void setCenterY(double centerY) {
    if (this.centerY != centerY) {
      this.centerY = centerY ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>y.</zh>
   * <en>Returns the coordinate of start y.</en>
   *
   * @return double <zh>y.</zh>
   * <en>The start y coordinate.</en>
   */
  public double getStartY() {
    return this.startY ;
  }

  /**
   * <zh>y,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the coordinate of start y,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param startY double <zh>y.</zh>
   * <en>The start y coordinate.</en>
   */
  public void setStartY(double startY) {
    if (this.startY != startY) {
      this.startY = startY ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>yյ.</zh>
   * <en>Returns the coordinate of end y.</en>
   *
   * @return double <zh>yյ.</zh>
   * <en>The end y coordinate.</en>
   */
  public double getEndY() {
    return this.endY ;
  }

  /**
   * <zh>yյ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the coordinate of end y,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param endY double <zh>yյ.</zh>
   * <en>The end y coordinate.</en>
   */
  public void setEndY(double endY) {
    if (this.endY != endY) {
      this.endY = endY ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ĵͼ.</zh>
   * <en>Returns the shape of center point.</en>
   *
   * @return Shape <zh>ͼζ (Ϊnull).</zh>
   * <en>A shape (maybe null).</en>
   */
  public Shape getCenterShape() {
    return this.centerShape ;
  }

  /**
   * <zh>ĵͼ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the shape of center point,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param shape Shape <zh>ͼζ (Ϊnull).</zh>
   * <en>A shape (null permitted).</en>
   */
  public void setCenterShape(Shape shape) {
    if (!(this.centerShape == null && shape == null)) {
      this.centerShape = shape ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>xyߵյĽߵĳ.</zh>
   * <en>Returns the length of the cap at the end of the error bars.</en>
   *
   * @return double <zh>.</zh>
   * <en>The length.</en>
   */
  public double getCapLength() {
    return this.capLength ;
  }

  /**
   * <zh>xyߵյĽߵĳ,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the length of the cap at the end of the error bars,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param length double <zh>,>=0.</zh>
   * <en>The length,must be >=0.</en>
   */
  public void setCapLength(double length) {
    if (length < 0) {
      throw new IllegalArgumentException("Argument 'length' must be >=0.") ;
    }

    if (this.capLength != length) {
      this.capLength = length ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ǷxError.</zh>
   * <en>Returns whether or not the x-error bars are drawn.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isDrawXError() {
    return this.drawXError ;
  }

  /**
   * <zh>ǷxError,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets whether or not the x-error bars are drawn,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setDrawXError(boolean flag) {
    if (this.drawXError != flag) {
      this.drawXError = flag ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ǷyError,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Returns whether or not the y-error bars are drawn,and sends a
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isDrawYError() {
    return this.drawYError ;
  }

  /**
   * <zh>ǷyError.</zh>
   * <en>Sets whether or not the y-error bars are drawn.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setDrawYError(boolean flag) {
    if (this.drawYError != flag) {
      this.drawYError = flag ;
      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼ豸(Ļӡ)ϻƴͼ</zh>
   * <en>Draw the ChartShape on a Java 2d graphics device(such as the screen or the printer).</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull).</zh>
   * <en>The graphics device (such as the screen or the printer) (null not permitted).</en>
   */
  public void draw(Graphics2D g2) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (!isVisible()) {
      return ;
    }

    Graphics2DState state = new Graphics2DState(g2) ;
    state.saveOldState() ;

    g2.translate(this.centerX, this.centerY) ;
    g2.rotate(getRotate()) ;
    beforeDraw(g2) ;

    float halfCapLen = (float) (this.capLength * 0.5) ;
    if (this.drawXError) { //x-error
      float leftX = (float) (this.startX - this.centerX) ;
      float rightX = (float) (this.endX - this.centerX) ;

      GeneralPath errorXPath = new GeneralPath() ;
      errorXPath.moveTo(leftX, halfCapLen) ;
      errorXPath.lineTo(leftX, -halfCapLen) ;
      errorXPath.lineTo(leftX, 0) ;
      errorXPath.lineTo(rightX, 0) ;
      errorXPath.lineTo(rightX, -halfCapLen) ;
      errorXPath.lineTo(rightX, halfCapLen) ;
      errorXPath.lineTo(rightX, 0) ;
      errorXPath.lineTo(leftX, 0) ;
      errorXPath.closePath() ;

      drawBackground(g2, errorXPath) ;
      drawBorder(g2, errorXPath) ;
    }

    if (this.drawYError) { //y-error
      float topY = (float) (this.startY - this.centerY) ;
      float bottomY = (float) (this.endY - this.centerY) ;
      GeneralPath errorYPath = new GeneralPath() ;
      errorYPath.moveTo(halfCapLen, topY) ;
      errorYPath.lineTo( -halfCapLen, topY) ;
      errorYPath.lineTo(0, topY) ;
      errorYPath.lineTo(0, bottomY) ;
      errorYPath.lineTo( -halfCapLen, bottomY) ;
      errorYPath.lineTo(halfCapLen, bottomY) ;
      errorYPath.lineTo(0, bottomY) ;
      errorYPath.lineTo(0, topY) ;
      errorYPath.closePath() ;

      drawBackground(g2, errorYPath) ;
      drawBorder(g2, errorYPath) ;
    }

    //мͼ
    if (this.centerShape != null) {
      drawBackground(g2, this.centerShape) ;
      drawBorder(g2, this.centerShape) ;
    }

    afterDraw(g2) ;
    g2.rotate( -getRotate()) ;
    state.restoreOldState() ;
    drawSelect(g2) ;
  }

  /**
   * <zh>ͼ豸(Ļӡ)ϻͼεѡв,Ϊͼοܻص,ԻͼʱҪȻеͼκ,
   * ٵô˷ѡв.</zh>
   * <en>Draw the selected ChartShape on a Java 2d graphics device(such
   * as the screen or the printer).</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(Ļӡ) (Ϊnull)</zh>
   * <en>the graphics device (such as the screen or the printer) (null not permitted).</en>
   */
  protected void drawSelect(Graphics2D g2) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }

    if (isSelected()) {
      Graphics2DState state = new Graphics2DState(g2) ;
      state.saveOldState() ;
      g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, getSelectRectAlpha())) ;

      double rectWidth = getSelectRectWidth() ;

      g2.setPaint(getSelectAnchorPaint()) ;
      g2.translate(this.centerX, this.centerY) ;
      g2.rotate(getRotate()) ;

      if (this.drawXError) {
        Point2D leftPoint = new Point2D.Double(this.startX - this.centerX, 0) ;
        Rectangle2D anchorLeft = AnchorUtil.getAnchorRect(leftPoint, rectWidth) ;
        g2.fill(anchorLeft) ;

        Point2D rightPoint = new Point2D.Double(this.endX - this.centerX, 0) ;
        Rectangle2D anchorRight = AnchorUtil.getAnchorRect(rightPoint, rectWidth) ;
        g2.fill(anchorRight) ;
      }

      if (this.drawYError) {
        Point2D topPoint = new Point2D.Double(0, this.startY - this.centerY) ;
        Rectangle2D anchorTop = AnchorUtil.getAnchorRect(topPoint, rectWidth) ;
        g2.fill(anchorTop) ;

        Point2D bottomPoint = new Point2D.Double(0, this.endY - this.centerY) ;
        Rectangle2D anchorBottom = AnchorUtil.getAnchorRect(bottomPoint, rectWidth) ;
        g2.fill(anchorBottom) ;
      }

      Point2D centerPoint = new Point2D.Double(0, 0) ;
      Rectangle2D anchorCenter = AnchorUtil.getAnchorRect(centerPoint, rectWidth) ;
      g2.fill(anchorCenter) ;

      g2.rotate( -getRotate()) ;
      state.restoreOldState() ;
    }

  }

  /**
   * <zh>ƶͼ,X᷽ƫtx,Y᷽ƫty,м
   * {@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Concatenates the current Graphics2D Transform with a translation transform.and sends
   * a {@link com.ijchart.xychart.event.ChartShapeChangeEvent} to all registered listeners.
   * Subsequent rendering is translated by the specified
   * distance relative to the previous position.
   * This is equivalent to calling transform(T), where T is an
   * AffineTransform represented by the following matrix:
   * <pre>
   *		[   1    0    tx  ]
   *		[   0    1    ty  ]
   *		[   0    0    1   ]
   * </pre>.</en>
   *
   * @param tx double <zh>X᷽ľ.</zh>
   * <en>The distance to translate along the x-axis.</en>
   * @param ty double <zh>Y᷽ľ.</zh>
   * <en>The distance to translate along the y-axis.</en>
   */
  public void translate(double tx, double ty) {
    if (!(tx == 0 && ty == 0)) {
      this.startX += tx ;
      this.endX += tx ;
      this.centerX += tx ;
      this.centerY += ty ;
      this.startY += ty ;
      this.endY += ty ;

      fireChartShapeChanged() ;
    }
  }

  /**
   * <zh>ͼϽǵ(ͼεת).</zh>
   * <en>Returns the top left corner coordinates of this shape (not consider rotation).</en>
   *
   * @return Point2D <zh>ֵ.</zh>
   * <en>The coordinate.</en>
   */
  public Point2D getTopLeftPoint() {
    double minX = Math.min(Math.min(this.startX, this.endX), this.centerX) ;
    double minY = Math.min(Math.min(this.startY, this.endY), this.centerY) ;

    return new Point2D.Double(minX, minY) ;
  }

  /**
   * <zh>ʱͼεĵ.</zh>
   * <en>Returns the lock coordinates of this shape.</en>
   *
   * @return Point2D <zh>ֵ.</zh>
   * <en>The coordinate.</en>
   */
  public Point2D getLockPoint() {
    return new Point2D.Double(this.centerX, this.centerY) ;
  }

  /**
   * <zh>任ͼ.</zh>
   * <en>Calculate the transformed shape.</en>
   */
  protected void calculateTransformedShape() {
    if (transformedShape == null) {
      float halfCapLen = (float) (this.capLength * 0.5) ;
      float leftX = (float) (this.startX - this.centerX) ;
      float rightX = (float) (this.endX - this.centerX) ;

      final float lineWidth = 1.0F ;
      GeneralPath transformShapePath = new GeneralPath() ;
      if (this.drawXError) { //x-error
        GeneralPath errorXPath = new GeneralPath() ;
        errorXPath.moveTo(leftX, halfCapLen) ;
        errorXPath.lineTo(leftX, -halfCapLen) ;
        errorXPath.lineTo(leftX, -lineWidth) ;
        errorXPath.lineTo(rightX, -lineWidth) ;
        errorXPath.lineTo(rightX, -halfCapLen) ;
        errorXPath.lineTo(rightX, halfCapLen) ;
        errorXPath.lineTo(rightX, lineWidth) ;
        errorXPath.lineTo(leftX, lineWidth) ;
        errorXPath.closePath() ;

        transformShapePath.append(errorXPath, false) ;
      }

      if (this.drawYError) { //y-error
        float topY = (float) (this.startY - this.centerY) ;
        float bottomY = (float) (this.endY - this.centerY) ;

        GeneralPath errorYPath = new GeneralPath() ;
        errorYPath.moveTo(halfCapLen, topY) ;
        errorYPath.lineTo( -halfCapLen, topY) ;
        errorYPath.lineTo( -lineWidth, topY) ;
        errorYPath.lineTo( -lineWidth, bottomY) ;
        errorYPath.lineTo( -halfCapLen, bottomY) ;
        errorYPath.lineTo(halfCapLen, bottomY) ;
        errorYPath.lineTo(lineWidth, bottomY) ;
        errorYPath.lineTo(lineWidth, topY) ;
        errorYPath.closePath() ;

        transformShapePath.append(errorYPath, false) ;
      }

      //мͼ
      if (this.centerShape != null) {
        transformShapePath.append(this.centerShape, false) ;
      } else {
        //ûͼ鹹һԲ,֤ѡ
        final double size = 8.0 ;
        final double delta = 4.0 ;
        Shape circle = new Ellipse2D.Double( -delta, -delta, size, size) ;

        transformShapePath.append(circle, false) ;
      }

      AffineTransform transform = new AffineTransform() ;
      transform.translate(this.centerX, this.centerY) ;
      transform.rotate(getRotate()) ;

      transformedShape = transform.createTransformedShape(transformShapePath) ;
    }
  }

  /**
   * <zh>ȵͼ.</zh>
   * <en>Returns the hotspot shape.</en>
   *
   * @return Shape <zh>ȵͼ (Ϊnull).</zh>
   * <en>The hotspot shape (maybe null).</en>
   */
  protected Shape getHotspotShape() {
    AffineTransform transform = AffineTransform.getTranslateInstance(this.centerX, this.centerY) ;
    transform.rotate(getRotate()) ;

    if (this.centerShape != null) {
      return transform.createTransformedShape(this.centerShape) ;
    } else {
      double minXWidth = Math.min(Math.abs(this.centerX - this.startX),
                                  Math.abs(this.endX - this.centerX)) ;
      double minYHeight = Math.min(Math.abs(this.centerY - this.startY),
                                   Math.abs(this.endY - this.centerY)) ;
      double rectWidth = Math.min(minXWidth, minYHeight) ;
      rectWidth = Math.max(2, rectWidth) ; //С2
      rectWidth = Math.min(rectWidth, 20) ; //20
      double halfRectWidth = 0.5 * rectWidth ;
      Rectangle2D rect = new Rectangle2D.Double( -halfRectWidth, -halfRectWidth,
                                                rectWidth, rectWidth) ;
      return transform.createTransformedShape(rect) ;
    }
  }

  /**
   * <zh>ͼε.</zh>
   * <en>The type for describe this shape.</en>
   *
   * @return ShapeType <zh>{@link ShapeType}ඨĳ (Ϊnull).</zh>
   * <en>The constant value in class {@link ShapeType}.</en>
   */
  public ShapeType getShapeType() {
    return ShapeType.SHAPE_ERROR_CROSS ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates whether some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument; false
   *  otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    if (!(obj instanceof ErrorCrossChartShape)) {
      return false ;
    }

    final double theta = 0.00001 ;
    ErrorCrossChartShape that = (ErrorCrossChartShape) obj ;
    if (!EqualsUtil.equals(this.startX, that.startX, theta)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.endX, that.endX, theta)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.centerX, that.centerX, theta)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.centerY, that.centerY, theta)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.startY, that.startY, theta)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.endY, that.endY, theta)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.centerShape, that.centerShape)) {
      return false ;
    }
    if (this.drawXError != that.drawXError) {
      return false ;
    }
    if (this.drawYError != that.drawYError) {
      return false ;
    }
    if (this.capLength != that.capLength) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double hashCode = super.hashCode() ;
    hashCode = 37 * hashCode + this.startX ;
    hashCode = 37 * hashCode + this.endX ;
    hashCode = 37 * hashCode + this.centerX ;
    hashCode = 37 * hashCode + this.centerY ;
    hashCode = 37 * hashCode + this.startY ;
    hashCode = 37 * hashCode + this.endY ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.centerShape) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.drawXError) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.drawYError) ;
    hashCode = 37 * hashCode + this.capLength ;

    return (int) hashCode ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object. </en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[startX=" + this.startX + ",endX=" + this.endX
      + ",centerX=" + this.centerX + ",centerY=" + this.centerY
      + ",startY=" + this.startY + ",endY=" + this.endY + "]" ;
  }

}
